const logger = require('../utils/oddUtils').getLogger();
const oddUtils = require('../utils/oddUtils');
const handleSnapshotCheckResult = require(
    './turnOrchestrator').handleSnapshotCheckResult;
const genericFetchUtils = require(
    '../utils/fillAndFetchUtils').genericFetchUtils;

const ALL_PROP_KEYS = oddUtils.getAllPropKeys();
const ALL_TRANSIENT_KEYS = oddUtils.getAllTransientKeys();
const CASES_SNAPSHOT_CHECK = oddUtils.casesForSnapshotMatchCheck();

const TURN_COOL_OFF = oddUtils.getTurnCoolOffPeriod();

function validateGameStateForTheTurn(game) {
  logger.debug({key: game.getId()},
      'No server validation available, passing the validation');
  return CASES_SNAPSHOT_CHECK.ALL_OK;
}

function handleReconExpiry(turnAndReconHolder, game) {
  const gameKey = game.getId();
  // check if turnReconHolder expired
  let curTime = Date.now();
  let turnExpiry = game.getCurrentTurnExpiry();

  if (turnAndReconHolder.reconExpiryTime < curTime) {
    logger.trace({key: gameKey}, 'turn recon holder', turnAndReconHolder);
    // what to do in case of recon expiry, need grace timers. here
    if (turnAndReconHolder.events['_tho']) {
      logger.debug({key: gameKey},
          'Turn handover is present but, reconcile event ' +
          'haven\'t reached, validating at server end.');
      let stateValidResult = validateGameStateForTheTurn(game);
      // pass on the result to act upon
      handleSnapshotCheckResult(stateValidResult, game);
    } else if (turnAndReconHolder.preForceReplayPeriod < curTime ||
        turnExpiry < curTime) {
      logger.debug({key: gameKey},
          'Turn hand over is not present, passing to grace time handler');
      checkGraceTimeAndHandle(game);
    }
  }
  // else nothing to do, lets wait
}

function checkGraceTimeAndHandle(game) {
  const gameKey = game.getId();
  let curPlrId = game.getCurrentTurnPlayerId();
  let curTime = Date.now();

  // set for the player that when he connects, force needs to be replayed
  game.setPlayerProp(curPlrId, ALL_PROP_KEYS.IS_FORCE_TO_BE_REPLAYED, true);

  // grace time can only kick in after turn is expired
  let turnExpiry = game.getCurrentTurnExpiry();
  if (turnExpiry < curTime) {
    logger.trace({key: gameKey},
        'Grace time kicked in, shaving time off grace time');
    // get when grace time started
    let graceTimeStart = parseInt(
        game.getPlayerProp(curPlrId, ALL_PROP_KEYS.GRACE_TIME_START)) ||
        turnExpiry;
    // shave grace time for the turn
    let graceTimeLeftForTurn =
        parseInt(game.getPlayerProp(curPlrId,
            ALL_PROP_KEYS.GRACE_TIME_LEFT_FOR_TURN)) -
        (curTime - graceTimeStart);
    // shave total grace time left
    let totalGraceTimeLeft = parseInt(
        game.getPlayerProp(curPlrId, ALL_PROP_KEYS.TOTAL_GRACE_TIME_LEFT)) -
        (curTime - graceTimeStart);

    // auto forfeit player, if grace time expires for the turn
    if (graceTimeLeftForTurn < 0) {
      logger.info({key: gameKey}, 'Grace time is over for the player', curPlrId,
          'for game',
          game.getId(), 'Auto forfeiting the match.');
      game.forfeitGameForPlayer(curPlrId, 'Grace time over.');
      return;
    }
    logger.trace({key: gameKey}, 'Grace time values for turn',
        graceTimeLeftForTurn,
        'total grace time left', totalGraceTimeLeft, 'grace time start',
        graceTimeStart);
    // set those values
    game.setPlayerProp(curPlrId, ALL_PROP_KEYS.GRACE_TIME_START, curTime);
    game.setPlayerProp(curPlrId, ALL_PROP_KEYS.GRACE_TIME_LEFT_FOR_TURN,
        graceTimeLeftForTurn);
    game.setPlayerProp(curPlrId, ALL_PROP_KEYS.TOTAL_GRACE_TIME_LEFT,
        totalGraceTimeLeft);
  }
}

function incrementIdleCountAndCheckAndForfeitMatch(game) {
  const gameKey = game.getId();
  let playerId = parseInt(game.getCurrentTurnPlayerId());

  let currentCount = parseInt(
      game.getPlayerProp(playerId, ALL_PROP_KEYS.IDLE_TURN_COUNT)) | 0;
  let meta = genericFetchUtils.fetchMeta(game);
  let allowedCount = meta.allowedIdleTurnsCount || 3;
  logger.trace({key: gameKey}, 'current turn count for player', playerId,
      currentCount);
  if (currentCount + 1 > allowedCount) {
    // if exceeded forfeit the match for the player
    logger.info({key: gameKey},
        'Exceeded the idle turnout time, forfeiting the match');
    game.forfeitGameForPlayer(playerId, 'Idle for too long');
    return true;
  }
  // just increment the count
  game.setPlayerProp(playerId, ALL_PROP_KEYS.IDLE_TURN_COUNT, currentCount + 1);
  return false;
}

function expireTurnIfRequired(turnExpiry, game) {
  const gameKey = game.getId();
  if (turnExpiry + TURN_COOL_OFF < Date.now()) {
    if (game.getGameProp(ALL_PROP_KEYS.RECONCILE_STATUS)) {
      // handling case when game got resumed from in-between state
      logger.info({key: gameKey},
          'Found already executed reconcile status in step function, ' +
          'acting upon it');
      handleSnapshotCheckResult(
          parseInt(game.getGameProp(ALL_PROP_KEYS.RECONCILE_STATUS)), game);
    } else {
      logger.info({key: gameKey}, 'Turn has expired switching for',
          game.getId());
      // send turn expired game event
      game.publishGameEvent(1001, [game.getCurrentTurnPlayerId()]);
      // idle turn count and match over check
      if (incrementIdleCountAndCheckAndForfeitMatch(game)) return;
      handleSnapshotCheckResult(validateGameStateForTheTurn(game), game);
    }
  }
}

const step = function(game) {
  const gameKey = game.getId();
  if (game.getState() >= 30)
    return logger.debug({key: gameKey}, 'Game has ended, nothing to do');

  let turnAndReconHolder = game.getTransient(
      ALL_TRANSIENT_KEYS.TURN_RECONCILE_HOLDER);
  if (oddUtils.parseBoolean(game.getGameProp(ALL_PROP_KEYS.IS_STRIKER_HIT)) &&
      !game.getGameProp(ALL_PROP_KEYS.RECONCILE_STATUS)) {
    return handleReconExpiry.call(this, turnAndReconHolder, game);
  }
  let currentTurnExpiry = game.getCurrentTurnExpiry();
  if (currentTurnExpiry > -1) {
    expireTurnIfRequired.call(this, currentTurnExpiry, game);
  }
};

module.exports = {step};