const oddUtils = require('../utils/oddUtils');
const logger = oddUtils.getLogger();
const turnFetchUtils = require('../utils/fillAndFetchUtils').turnFetchUtils;
const genericFetchUtils = require(
    '../utils/fillAndFetchUtils').genericFetchUtils;
const hangman = require('./hangman');
const lifeCycle = require('./lifecycle');
const TurnAndReconcileStateHolder = require('../utils/turnAndReconHolder');
const {_getGameProp,_setGameProp,_getPlayerProp,_setPlayerProp} =
    require('../utils/utils')
const ENUMS = require("../utils/enums")
const ALL_PROP_KEYS = oddUtils.getAllPropKeys();
const ALL_PROP_VALUES = oddUtils.getALLPropValues();
const ALL_TRANSIENT_KEYS = oddUtils.getAllTransientKeys();
const CASES_SNAPSHOT_CHECK = oddUtils.casesForSnapshotMatchCheck();

const processTurn = function(game) {
  // player and turn info
  const gameKey = game.getId();
  let playerId = parseInt(game.getCurrentTurnPlayerId());
  let currentTurnIndex = parseInt(game.getCurrentTurnIndex());
  let currentTack = parseInt(game.getGameProp(ALL_PROP_KEYS.POTT_TACK));

  // params that decide cases
  let pottedCoins = turnFetchUtils.fetchCoinsForTurnAndTack(game,
      currentTurnIndex, currentTack);
  if(currentTurnIndex === 1){
    game.publishGameEvent(1010, [playerId],pottedCoins.length);
  }
  logger.debug({key: gameKey}, 'List of potted coins are',
      JSON.stringify(pottedCoins));
  let isRedPottedNow = isRedPottedForTack(game, currentTurnIndex, currentTack);
  let isRedPottedBefore = isRedPottedForTack(game, currentTurnIndex,
      currentTack - 1);
  let isStrikerPotted = isStrikerPottedForTack(game);

  let caseToHandle =
      detectCaseToHandle(game, playerId, pottedCoins, isRedPottedNow,
          isRedPottedBefore, isStrikerPotted);
  caseToHandle.isRedPottedNow = isRedPottedNow;
  logger.debug({key: gameKey}, 'Detected case ', caseToHandle, 'for player',
      playerId, 'turn',
      currentTurnIndex, 'tack', currentTack);

  let judgment = getJudgment(game, caseToHandle, playerId);

  function clientCommunication(res) { // to be executed after ack and state is saved
    // pass judgment to players
    rewardOrPunishPlayers(game, judgment, playerId);

    // send properties of concerned coins to client
    let objIdsToUpdate = new Set();
    if (isRedPottedBefore || isRedPottedNow)
      objIdsToUpdate.add(genericFetchUtils.fetchRedCoin(game).id);
    for (let coin of caseToHandle.coinsToAwardOpp)
      objIdsToUpdate.add(coin.id);
    for (let coin of caseToHandle.coinsToPutBack)
      objIdsToUpdate.add(coin.id);
    for (let coin of caseToHandle.coinsToAwardSelf)
      objIdsToUpdate.add(coin.id);
    game.syncGameObjServVars(...objIdsToUpdate.values());

    // sync player vars
    if (isRedPottedNow || isRedPottedBefore ||
        lifeCycle.grantGraceTimeForTurn(game, playerId))
      game.syncPlayerServVars(playerId);

    game.setGameProp(ALL_PROP_KEYS.IS_STRIKER_HIT, false);
    game.syncGameStateServVars();

    // send force update of positions (if any) to players
    if (typeof res === 'function') {
      logger.debug({key: gameKey}, 'Sending force update of coins to clients');
      res();
    }
  }

  // check if gameEnded
  let shouldGameBeEnded = lifeCycle.shouldGameBeEnded(game, judgment, playerId);

  if (shouldGameBeEnded) {
    // award score and sync updated props to client before sending end game message
    clientCommunication();
    lifeCycle.endGame(game, oddUtils.getKOWinType());
    return;
  }

  // put back coins and send the force update to client
  hangman.guillotine(game, caseToHandle.coinsToPutBack, (err, res) => {
    if (err) {
      logger.error({key: gameKey},
          'Error in putting back coins to centre of the board', err);
    }

    // sending to clients the required data changes (score, coin properties change)
    clientCommunication(res);
    game.save(() => {
      if (err) return logger.error({key: gameKey}, 'Error in saving gamestate,',
          err);

      // communicate with neutrino should turn continue and foul after coins are put back
      logger.debug({key: gameKey}, 'Passing end turn result to neutrino');
      if (caseToHandle.isTurnContinue)
        lifeCycle.continueTurn(game, caseToHandle.isFoul);
      else lifeCycle.nextTurn(game, caseToHandle.isFoul);
    });
  });
};

function isRedPottedForTack(game, turnIndex, tack) {
  let redCoin = genericFetchUtils.fetchRedCoin(game);
  let isPotted = oddUtils.parseBoolean(
      redCoin.getProp(ALL_PROP_KEYS.IS_POTTED));
  let isRedScoreAwarded = oddUtils.parseBoolean(
      redCoin.getProp(ALL_PROP_KEYS.IS_SCORED));
  if (isPotted && !isRedScoreAwarded) {
    // if red is already awarded then no need to say it was potted
    let coinTurnIndex = parseInt(redCoin.getProp(ALL_PROP_KEYS.POTT_TURN));
    let coinTack = parseInt(redCoin.getProp(ALL_PROP_KEYS.POTT_TACK));
    return turnIndex === coinTurnIndex && tack === coinTack;
  }
  return false;
}

function isStrikerPottedForTack(game) {
  let striker = genericFetchUtils.fetchStriker(game);
  return oddUtils.parseBoolean(striker.getProp(ALL_PROP_KEYS.IS_POTTED));
}

function getJudgment(game, caseToHandle, playerId) {
  let redCoin = genericFetchUtils.fetchRedCoin(game);

  let isRedScored = false;
  let selfScore = 0;
  let oppScore = 0;

  for (let coin of caseToHandle.coinsToAwardSelf) {
    coin.setProp(ALL_PROP_KEYS.IS_SCORED, true);
    if (coin === redCoin)
      isRedScored = true; // if red is scored, dont increase the score just mark it
    else
      selfScore++; // if a coin is scored then mark it
  }

  // opponent coins increase opponent score
  for (let coin of caseToHandle.coinsToAwardOpp) {
    oppScore++;
    coin.setProp(ALL_PROP_KEYS.IS_SCORED, true);
  }

  // for coins to put back, if plucked from already scored coin, then decrease self score
  for (let coin of caseToHandle.coinsToPutBack) {
    if (oddUtils.getColorOfCoinByType(coin.type) ===
        genericFetchUtils.fetchPlayerColor(game, playerId) &&
        oddUtils.parseBoolean(coin.getProp(ALL_PROP_KEYS.IS_SCORED)))
      selfScore--;
    // if red should be put back remove pocketed against player
    if (coin === redCoin)
      game.setPlayerProp(playerId, ALL_PROP_KEYS.DID_POCKETED_RED, false);

    coin.setProp(ALL_PROP_KEYS.IS_POTTED, false);
    coin.setProp(ALL_PROP_KEYS.IS_SCORED, false);
    coin.setProp(ALL_PROP_KEYS.POTT_TURN, -1);
    coin.setProp(ALL_PROP_KEYS.POTT_TACK, -1);
  }

  // if red is scored mark against the user and redCoin
  if (isRedScored) {
    game.setPlayerProp(playerId, ALL_PROP_KEYS.DID_SCORED_RED, true);
    redCoin.setProp(ALL_PROP_KEYS.IS_SCORED, true);
  }

  // if red is pocketed and turn to be continued then against player set red as pocketed
  if (caseToHandle.isRedPottedNow && caseToHandle.isTurnContinue)
    game.setPlayerProp(playerId, ALL_PROP_KEYS.DID_POCKETED_RED, true);

  return {isRedScored: isRedScored, oppScore: oppScore, selfScore: selfScore};
}

function rewardOrPunishPlayers(game, judgment, playerId) {
  let selfScore = judgment.selfScore;
  let oppScore = judgment.oppScore;
  let opponentId = genericFetchUtils.fetchOpponentId(game, playerId);

  // award scores to player and opponent
  if (selfScore !== 0)
    game.awardScore(playerId, selfScore);
  if (oppScore !== 0) {
    _setPlayerProp(game,playerId,ENUMS.PLAYER_PROPS.PLAYER_WRONG_COIN_POCKETED_COUNT,
        _getPlayerProp(game,playerId,ENUMS.PLAYER_PROPS.PLAYER_WRONG_COIN_POCKETED_COUNT)+1)
    game.awardScore(opponentId, oppScore);
  }

}

/**
 * Game rules logic
 * @returns {{coinsToAwardSelf: *, isTurnContinue: *, isFoul: *, coinsToAwardOpp: *, coinsToPutBack: *}}
 */
function detectCaseToHandle(
    game, playerId, pottedCoins, isRedPottedNow, isRedPottedBefore,
    isStrikerPotted) {
  const gameKey = game.getId();
  let coinsToPutBack = new Set();
  let coinsToAwardSelf = new Set();
  let coinsToAwardOpp = new Set();
  let isTurnContinue = false;
  let isFoul = false;
  logger.trace({key: gameKey}, 'Printing fillfetch', game.fillFetch);

  // player params
  let playerColor = genericFetchUtils.fetchPlayerColor(game, playerId);
  let playerScore = game.getPlayerScore(playerId);
  // opponent params
  let oppPlayerId = genericFetchUtils.fetchOpponentId(game, playerId);
  let oppScore = game.getPlayerScore(oppPlayerId);
  let oppColor = genericFetchUtils.fetchPlayerColor(game, oppPlayerId);
  // useful objs/constants
  let redCoin = genericFetchUtils.fetchRedCoin(game);
  let maxGameEndScore = genericFetchUtils.fetchGameEndScore(game);

  // markers to act up on
  let shouldRedPutBack = false;
  let timesShouldBePunished = 0;
  let timesAlreadyPunished = 0;
  let isFollowUpHappened = false;
  let shouldOppCoinBePutBack = false;
  let isOppCoinAlreadyPutBack = false;
  let pottedOppCoins=false;

  /*
   * #### Start of RULES ####
   */
  // last coin foul detection
  let isRedAlreadyScored = oddUtils.parseBoolean(
      redCoin.getProp(ALL_PROP_KEYS.IS_SCORED));
  let countSelfCoinsScored = 0;
  let countOppCoinsScored = 0;
  for (let coin of pottedCoins)
    if (oddUtils.getColorOfCoinByType(coin.type) === playerColor) {
      countSelfCoinsScored++;
      isFollowUpHappened = true;
    }
    else if (oddUtils.getColorOfCoinByType(coin.type) === oppColor) {
      countOppCoinsScored++;
      pottedOppCoins  = true
    }

  // foul detection, cases
  // a) when striker is potted
  // b) when last coin of self is potted and red is not scored yet
  // c) when last coin of opp is potted and red is not scored yet
  if (isStrikerPotted) {
    isFoul = true;
    timesShouldBePunished = 1;
  }

  if (!isRedAlreadyScored && playerScore + countSelfCoinsScored >=
      maxGameEndScore &&
      !((isRedPottedNow || isRedPottedBefore) && isFollowUpHappened &&
          !isStrikerPotted)) { // case B should be punished twice
    isFoul = true;
    if(playerColor === ALL_PROP_VALUES.COLOR.WHITE){
      game.sendTip(playerId,4)
    }else {
      game.sendTip(playerId,3)
    }
    timesShouldBePunished = 2;
  }
  if (!isRedAlreadyScored
      && oppScore + countOppCoinsScored >= maxGameEndScore &&
      !((isRedPottedNow || isRedPottedBefore) && isFollowUpHappened &&
          !isStrikerPotted)) { // case C should be punished twice
    isFoul = true;
    timesShouldBePunished = 2;
    // mark opp coin to put back as true, if opp last coin is potted without red being scored by anyone
    shouldOppCoinBePutBack = true;
  }

  // now run through all potted coins and decide what to award whom
  for (let coin of pottedCoins) {
    // opponent potted coin cases
    if (oddUtils.getColorOfCoinByType(coin.type) === oppColor) {
      // opp last coins are potted without red being scored then one opp coin needs to be put back on
      if (shouldOppCoinBePutBack && !isOppCoinAlreadyPutBack) {
        coinsToPutBack.add(coin);
        isOppCoinAlreadyPutBack = true; // mark as opp coin is already put back
      } else {
        coinsToAwardOpp.add(coin);
      }
    }
    // self potted coin cases
    else if (oddUtils.getColorOfCoinByType(coin.type) === playerColor) {
      // put back coin when there is foul and not punished yet
      if (isFoul && timesShouldBePunished > timesAlreadyPunished) {
        coinsToPutBack.add(coin);
        timesAlreadyPunished++; // increment number of times punished
      } else { // if no foul or already punished then award the coin (multi coins case)
        coinsToAwardSelf.add(coin);
      }
    }
  }

  // check if red should be awarded, if should be awarded then award it
  if ((isRedPottedNow || isRedPottedBefore) && isFollowUpHappened && !isFoul) {
    _setPlayerProp(game,playerId,ENUMS.PLAYER_PROPS.IS_PLAYER_POCKETED_RED,true)
    coinsToAwardSelf.add(redCoin); // award red coin to player
  }

  // if there is foul red should be put back
  // and if no followup happened red should be put back
  if ((isFoul && (isRedPottedBefore || isRedPottedNow)) ||
      (isRedPottedBefore && !isFollowUpHappened)){
    shouldRedPutBack = true;
    }
  if(isRedPottedBefore && !isFollowUpHappened){
    if(playerColor === ALL_PROP_VALUES.COLOR.WHITE){
      game.sendTip(playerId,6)
    }else {
      game.sendTip(playerId,5)
    }
  }
  // if foul happened but self is not punished then punish till count matches
  while (isFoul && timesShouldBePunished > timesAlreadyPunished) {
    let coinFetched = fetchAScoredCoin(game, playerColor);
    if (coinFetched) coinsToPutBack.add(coinFetched);
    timesAlreadyPunished++;
  }

  // if red should be put back, add it to coinsToPutBack
  if (shouldRedPutBack)
    coinsToPutBack.add(genericFetchUtils.fetchRedCoin(game));

  // decide turn continue
  if (!isFoul && (isFollowUpHappened || isRedPottedNow))
    isTurnContinue = true;
  if(pottedOppCoins && !isFollowUpHappened){
    logger.trace({key: gameKey}, 'Printing is tip sent',playerColor);
    if(playerColor === ALL_PROP_VALUES.COLOR.WHITE){
      game.sendTip(playerId,1)
    }else {
      game.sendTip(playerId,2)
    }
  }
  logger.info({key: gameKey}, 'The case detected is', JSON.stringify({
      shouldRedPutBack,
      timesShouldBePunished,
      timesAlreadyPunished,
      isFollowUpHappened,
      shouldOppCoinBePutBack,
      isOppCoinAlreadyPutBack,
      playerScore,
      oppScore,
      maxGameEndScore,
      pottedCoins,
      isRedAlreadyScored,
      isRedPottedNow,
      isRedPottedBefore,
      isStrikerPotted,
      isTurnContinue,
      coinsToPutBack,
      coinsToAwardSelf,
      coinsToAwardOpp,
    }));

  return {
    isFoul: isFoul,
    coinsToPutBack: [...coinsToPutBack],
    coinsToAwardSelf: [...coinsToAwardSelf],
    coinsToAwardOpp: [...coinsToAwardOpp],
    isTurnContinue: isTurnContinue,
  };
}

function fetchAScoredCoin(game, color) {
  let coinsToLookUp;
  // if color is white fetch white coins or if color is black fetch black coins
  if (color === ALL_PROP_VALUES.COLOR.WHITE)
    coinsToLookUp = genericFetchUtils.fetchWhiteCoins(game);
  else if (color === ALL_PROP_VALUES.COLOR.BLACK)
    coinsToLookUp = genericFetchUtils.fetchBlackCoins(game);
  if (!coinsToLookUp) throw new Error('Unknown color supplied: ' + color);

  // iterate over coins to fetch if there is any potted coin, if present return the potted coin object
  for (let coin of coinsToLookUp)
    if (oddUtils.parseBoolean(coin.getProp(ALL_PROP_KEYS.IS_SCORED)))
      return coin;
}

function fetchTurnReconHolder(game) {
  let turnReconHolder = game.getTransient(
      ALL_TRANSIENT_KEYS.TURN_RECONCILE_HOLDER);
  if (!turnReconHolder) {
    turnReconHolder = new TurnAndReconcileStateHolder();
    game.addTransient(ALL_TRANSIENT_KEYS.TURN_RECONCILE_HOLDER,
        turnReconHolder);
  }
  return turnReconHolder;
}

const handleSnapshotCheckResult = function(
    stateMatchResult, game, gameObjectsToCheck) {
  const gameKey = game.getId();
  logger.trace({key: gameKey}, 'Handling checkResult', stateMatchResult);
  if (stateMatchResult === CASES_SNAPSHOT_CHECK.ALL_OK) {
    if (gameObjectsToCheck) {
      game.forceUpdateGameObjects(...gameObjectsToCheck);
    }
    processTurn(game);
  } else if (stateMatchResult === CASES_SNAPSHOT_CHECK.ABANDON_MATCH) {
    lifeCycle.abandonMatch(game);
  } else if (stateMatchResult ===
      CASES_SNAPSHOT_CHECK.REVERT_TO_PREVIOUS_SAVEPOINT) {
    lifeCycle.revertToPreviousSavePoint(game);
  }
};

function checkAndProcess(game) {
  const gameKey = game.getId();
  let turnReconHolder = fetchTurnReconHolder(game);
  if (turnReconHolder.canComplete()) {
    logger.debug({key: gameKey},
        'Received both turn handover & recon, will process the turn');
    let gameObjectsToCheck = turnReconHolder.events['_rc'];
    process.nextTick(() => {
      let sateMatchResult = hangman.trialByPosition(game, gameObjectsToCheck);
      game.setGameProp(ALL_PROP_KEYS.RECONCILE_STATUS, sateMatchResult);
      game.save(() => {
        handleSnapshotCheckResult(sateMatchResult, game, gameObjectsToCheck);
      });
    });
  }

}

const handOverTurn = function(game, cb) {
  const gameKey = game.getId();
  logger.debug({key: gameKey}, 'Received turn handover for turn:',
      game.getCurrentTurnIndex(),
      'tack:', game.getGameProp(ALL_PROP_KEYS.POTT_TACK));
  let turnReconHolder = fetchTurnReconHolder(game);
  turnReconHolder.addTurnHandoverEvent({});
  // handover turn is received reset player ide count to zero
  let playerId = parseInt(game.getCurrentTurnPlayerId());
  game.setPlayerProp(playerId, ALL_PROP_KEYS.IDLE_TURN_COUNT, 0);
  // send no-op as we need to validate
  cb({isTurnContinue: 'NO-OP'});
  checkAndProcess(game);
};

const postPositionReconcileHandler = function(game, gameObjects) {
  const gameKey = game.getId();
  let serverElapsed = game.getGameStateTimeElapsed();
  for (let g of gameObjects) if (g.timeElapsed < serverElapsed)
    return logger.warn({key: gameKey},
        'Received stale reconcile event, dropping it');
  logger.debug({key: gameKey}, 'Received post position recon for turn:',
      game.getCurrentTurnIndex(), 'tack:',
      game.getGameProp(ALL_PROP_KEYS.POTT_TACK));
  let turnReconHolder = fetchTurnReconHolder(game);
  turnReconHolder.addReconcileValidateEvent(gameObjects);
  checkAndProcess(game);
};

module.exports = {
  handleSnapshotCheckResult,
  handOverTurn,
  postPositionReconcileHandler,
};